@@ -1,8 +1,12 @@
 use crate::types::{Asn1Readable, SimpleAsn1Readable, Tlv};
+use alloc::borrow::Cow;
+use alloc::string::ToString;
+use alloc::vec::Vec;
+use core::fmt;
 
 /// ParseError are returned when there is an error parsing the ASN.1 data.
 #[derive(Debug, PartialEq)]
-pub enum ParseError {
+pub enum ParseErrorKind {
     /// Something about the value was invalid.
     InvalidValue,
     /// An unexpected tag was encountered.
@@ -19,6 +23,60 @@ pub enum ParseError {
     EncodedDefault,
 }
 
+#[derive(Debug, PartialEq)]
+#[doc(hidden)]
+pub enum ParseLocation {
+    Field(&'static str),
+    Index(usize),
+}
+
+#[derive(PartialEq)]
+pub struct ParseError {
+    kind: ParseErrorKind,
+    parse_locations: [Option<ParseLocation>; 8],
+    parse_depth: u8,
+}
+
+impl ParseError {
+    pub fn new(kind: ParseErrorKind) -> ParseError {
+        ParseError {
+            kind,
+            parse_locations: [None, None, None, None, None, None, None, None],
+            parse_depth: 0,
+        }
+    }
+
+    #[doc(hidden)]
+    pub fn add_location(mut self, loc: ParseLocation) -> Self {
+        if (self.parse_depth as usize) < self.parse_locations.len() {
+            self.parse_locations[self.parse_depth as usize] = Some(loc);
+            self.parse_depth += 1;
+        }
+        self
+    }
+}
+
+impl fmt::Debug for ParseError {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        let mut f = f.debug_struct("ParseError");
+        f.field("kind", &self.kind);
+        if self.parse_depth > 0 {
+            f.field(
+                "location",
+                &self.parse_locations[..self.parse_depth as usize]
+                    .iter()
+                    .rev()
+                    .map(|l| match l.as_ref().unwrap() {
+                        ParseLocation::Field(f) => Cow::<'_, str>::Borrowed(f),
+                        ParseLocation::Index(i) => Cow::<'_, str>::Owned(i.to_string()),
+                    })
+                    .collect::<Vec<_>>(),
+            );
+        }
+        f.finish()
+    }
+}
+
 /// The result of a `parse`. Either a successful value or a `ParseError`.
 pub type ParseResult<T> = Result<T, ParseError>;
 
@@ -56,7 +114,7 @@ impl<'a> Parser<'a> {
     #[inline]
     fn finish(self) -> ParseResult<()> {
         if !self.is_empty() {
-            return Err(ParseError::ExtraData);
+            return Err(ParseError::new(ParseErrorKind::ExtraData));
         }
         Ok(())
     }
@@ -72,7 +130,7 @@ impl<'a> Parser<'a> {
     #[inline]
     fn read_u8(&mut self) -> ParseResult<u8> {
         if self.data.is_empty() {
-            return Err(ParseError::ShortData);
+            return Err(ParseError::new(ParseErrorKind::ShortData));
         }
         let (val, data) = self.data.split_at(1);
         self.data = data;
@@ -82,7 +140,7 @@ impl<'a> Parser<'a> {
     #[inline]
     fn read_bytes(&mut self, length: usize) -> ParseResult<&'a [u8]> {
         if length > self.data.len() {
-            return Err(ParseError::ShortData);
+            return Err(ParseError::new(ParseErrorKind::ShortData));
         }
         let (result, data) = self.data.split_at(length);
         self.data = data;
@@ -97,25 +155,25 @@ impl<'a> Parser<'a> {
         let num_bytes = b & 0x7f;
         // Indefinite length form is not valid DER
         if num_bytes == 0 {
-            return Err(ParseError::InvalidValue);
+            return Err(ParseError::new(ParseErrorKind::InvalidValue));
         }
 
         let mut length = 0;
         for _ in 0..num_bytes {
             let b = self.read_u8()?;
             if length > (usize::max_value() >> 8) {
-                return Err(ParseError::IntegerOverflow);
+                return Err(ParseError::new(ParseErrorKind::IntegerOverflow));
             }
             length <<= 8;
             length |= b as usize;
             // Disallow leading 0s
             if length == 0 {
-                return Err(ParseError::InvalidValue);
+                return Err(ParseError::new(ParseErrorKind::InvalidValue));
             }
         }
         // Do not allow values <0x80 to be encoded using the long form
         if length < 0x80 {
-            return Err(ParseError::InvalidValue);
+            return Err(ParseError::new(ParseErrorKind::InvalidValue));
         }
         Ok(length)
     }
@@ -184,8 +242,8 @@ mod tests {
     use crate::types::Asn1Readable;
     use crate::{
         BigUint, BitString, Choice1, Choice2, Choice3, Enumerated, GeneralizedTime, IA5String,
-        ObjectIdentifier, ParseError, ParseResult, PrintableString, Sequence, SequenceOf, SetOf,
-        Tlv, UtcTime, Utf8String, VisibleString,
+        ObjectIdentifier, ParseError, ParseErrorKind, ParseLocation, ParseResult, PrintableString,
+        Sequence, SequenceOf, SetOf, Tlv, UtcTime, Utf8String, VisibleString,
     };
     #[cfg(feature = "const-generics")]
     use crate::{Explicit, Implicit};
@@ -228,6 +286,25 @@ mod tests {
         assert_eq!(result, b"\x00");
     }
 
+    #[test]
+    fn test_parse_error_debug() {
+        for (e, expected) in [
+            (
+                ParseError::new(ParseErrorKind::InvalidValue),
+                "ParseError { kind: InvalidValue }",
+            ),
+            (
+                ParseError::new(ParseErrorKind::InvalidValue)
+                    .add_location(ParseLocation::Field("Abc::123")),
+                "ParseError { kind: InvalidValue, location: [\"Abc::123\"] }",
+            ),
+        ]
+        .iter()
+        {
+            assert_eq!(&format!("{:?}", e), expected)
+        }
+    }
+
     fn assert_parses_cb<
         'a,
         T: fmt::Debug + PartialEq,
@@ -239,7 +316,7 @@ mod tests {
     ) {
         for (expected, der_bytes) in data {
             let result = crate::parse(der_bytes, &f);
-            assert_eq!(&result, expected)
+            assert_eq!(&result, expected);
         }
     }
 
@@ -253,7 +330,7 @@ mod tests {
     #[test]
     fn test_parse_extra_data() {
         let result = crate::parse(b"\x00", |_| Ok(()));
-        assert_eq!(result, Err(ParseError::ExtraData));
+        assert_eq!(result, Err(ParseError::new(ParseErrorKind::ExtraData)));
     }
 
     #[test]
@@ -273,7 +350,10 @@ mod tests {
         assert_parses_cb(
             &[
                 (Ok(8), b"\x02\x01\x08"),
-                (Err(E::P(ParseError::ShortData)), b"\x02\x01"),
+                (
+                    Err(E::P(ParseError::new(ParseErrorKind::ShortData))),
+                    b"\x02\x01",
+                ),
                 (Err(E::X(7)), b"\x02\x01\x07"),
             ],
             |p| {
@@ -298,9 +378,12 @@ mod tests {
                 }),
                 b"\x04\x03abc",
             ),
-            (Err(ParseError::ShortData), b"\x04\x03a"),
-            (Err(ParseError::ShortData), b"\x04"),
-            (Err(ParseError::ShortData), b""),
+            (
+                Err(ParseError::new(ParseErrorKind::ShortData)),
+                b"\x04\x03a",
+            ),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x04"),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b""),
         ]);
     }
 
@@ -308,7 +391,10 @@ mod tests {
     fn test_parse_null() {
         assert_parses::<()>(&[
             (Ok(()), b"\x05\x00"),
-            (Err(ParseError::InvalidValue), b"\x05\x01\x00"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x05\x01\x00",
+            ),
         ]);
     }
 
@@ -317,10 +403,22 @@ mod tests {
         assert_parses::<bool>(&[
             (Ok(true), b"\x01\x01\xff"),
             (Ok(false), b"\x01\x01\x00"),
-            (Err(ParseError::InvalidValue), b"\x01\x00"),
-            (Err(ParseError::InvalidValue), b"\x01\x01\x01"),
-            (Err(ParseError::InvalidValue), b"\x01\x02\x00\x00"),
-            (Err(ParseError::InvalidValue), b"\x01\x02\xff\x01"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x01\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x01\x01\x01",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x01\x02\x00\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x01\x02\xff\x01",
+            ),
         ]);
     }
 
@@ -333,15 +431,15 @@ mod tests {
                 Ok(b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
                 b"\x04\x81\x81aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
             ),
-            (Err(ParseError::InvalidValue), b"\x04\x80"),
-            (Err(ParseError::InvalidValue), b"\x04\x81\x00"),
-            (Err(ParseError::InvalidValue), b"\x04\x81\x01\x09"),
+            (Err(ParseError::new(ParseErrorKind::InvalidValue)), b"\x04\x80"),
+            (Err(ParseError::new(ParseErrorKind::InvalidValue)), b"\x04\x81\x00"),
+            (Err(ParseError::new(ParseErrorKind::InvalidValue)), b"\x04\x81\x01\x09"),
             (
-                Err(ParseError::IntegerOverflow),
+                Err(ParseError::new(ParseErrorKind::IntegerOverflow)),
                 b"\x04\x89\x01\x01\x01\x01\x01\x01\x01\x01\x01"
             ),
-            (Err(ParseError::ShortData), b"\x04\x03\x01\x02"),
-            (Err(ParseError::ShortData), b"\x04\x83\xff\xff\xff\xff\xff\xff"),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x04\x03\x01\x02"),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x04\x83\xff\xff\xff\xff\xff\xff"),
         ]);
     }
 
@@ -359,20 +457,34 @@ mod tests {
                 Ok(core::i64::MAX),
                 b"\x02\x08\x7f\xff\xff\xff\xff\xff\xff\xff",
             ),
-            (Err(ParseError::UnexpectedTag { actual: 0x3 }), b"\x03\x00"),
-            (Err(ParseError::ShortData), b"\x02\x02\x00"),
-            (Err(ParseError::ShortData), b""),
-            (Err(ParseError::ShortData), b"\x02"),
             (
-                Err(ParseError::IntegerOverflow),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x3,
+                })),
+                b"\x03\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::ShortData)),
+                b"\x02\x02\x00",
+            ),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b""),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x02"),
+            (
+                Err(ParseError::new(ParseErrorKind::IntegerOverflow)),
                 b"\x02\x09\x02\x00\x00\x00\x00\x00\x00\x00\x00",
             ),
             (
-                Err(ParseError::InvalidValue),
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
                 b"\x02\x05\x00\x00\x00\x00\x01",
             ),
-            (Err(ParseError::InvalidValue), b"\x02\x02\xff\x80"),
-            (Err(ParseError::InvalidValue), b"\x02\x00"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x02\xff\x80",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x00",
+            ),
         ])
     }
 
@@ -383,9 +495,12 @@ mod tests {
                 Ok(core::u64::MAX),
                 b"\x02\x09\x00\xff\xff\xff\xff\xff\xff\xff\xff",
             ),
-            (Err(ParseError::InvalidValue), b"\x02\x01\xff"),
             (
-                Err(ParseError::IntegerOverflow),
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x01\xff",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::IntegerOverflow)),
                 b"\x02\x09\x02\x00\x00\x00\x00\x00\x00\x00\x00",
             ),
         ]);
@@ -397,8 +512,14 @@ mod tests {
             (Ok(0i8), b"\x02\x01\x00"),
             (Ok(127i8), b"\x02\x01\x7f"),
             (Ok(-128i8), b"\x02\x01\x80"),
-            (Err(ParseError::IntegerOverflow), b"\x02\x02\x02\x00"),
-            (Err(ParseError::InvalidValue), b"\x02\x00"),
+            (
+                Err(ParseError::new(ParseErrorKind::IntegerOverflow)),
+                b"\x02\x02\x02\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x00",
+            ),
         ])
     }
 
@@ -408,8 +529,14 @@ mod tests {
             (Ok(0u8), b"\x02\x01\x00"),
             (Ok(127u8), b"\x02\x01\x7f"),
             (Ok(255u8), b"\x02\x02\x00\xff"),
-            (Err(ParseError::IntegerOverflow), b"\x02\x02\x01\x00"),
-            (Err(ParseError::InvalidValue), b"\x02\x01\x80"),
+            (
+                Err(ParseError::new(ParseErrorKind::IntegerOverflow)),
+                b"\x02\x02\x01\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x01\x80",
+            ),
         ])
     }
 
@@ -422,9 +549,18 @@ mod tests {
                 Ok(BigUint::new(b"\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff").unwrap()),
                 b"\x02\x0d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff",
             ),
-            (Err(ParseError::InvalidValue), b"\x02\x00"),
-            (Err(ParseError::InvalidValue), b"\x02\x01\x80"),
-            (Err(ParseError::InvalidValue), b"\x02\x02\xff\x80"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x01\x80",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x02\xff\x80",
+            ),
         ]);
     }
 
@@ -455,12 +591,18 @@ mod tests {
                 Ok(ObjectIdentifier::from_string("2.100.3").unwrap()),
                 b"\x06\x03\x81\x34\x03",
             ),
-            (Err(ParseError::InvalidValue), b"\x06\x00"),
             (
-                Err(ParseError::InvalidValue),
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x06\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
                 b"\x06\x07\x55\x02\xc0\x80\x80\x80\x80",
             ),
-            (Err(ParseError::InvalidValue), b"\x06\x02\x2a\x86"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x06\x02\x2a\x86",
+            ),
         ])
     }
 
@@ -474,10 +616,22 @@ mod tests {
                 Ok(BitString::new(b"\x81\xf0", 4).unwrap()),
                 b"\x03\x03\x04\x81\xf0",
             ),
-            (Err(ParseError::InvalidValue), b"\x03\x00"),
-            (Err(ParseError::InvalidValue), b"\x03\x02\x07\x01"),
-            (Err(ParseError::InvalidValue), b"\x03\x02\x07\x40"),
-            (Err(ParseError::InvalidValue), b"\x03\x02\x08\x00"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x03\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x03\x02\x07\x01",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x03\x02\x07\x40",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x03\x02\x08\x00",
+            ),
         ]);
     }
 
@@ -486,7 +640,10 @@ mod tests {
         assert_parses::<PrintableString>(&[
             (Ok(PrintableString::new("abc").unwrap()), b"\x13\x03abc"),
             (Ok(PrintableString::new(")").unwrap()), b"\x13\x01)"),
-            (Err(ParseError::InvalidValue), b"\x13\x03ab\x00"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x13\x03ab\x00",
+            ),
         ]);
     }
 
@@ -495,7 +652,10 @@ mod tests {
         assert_parses::<IA5String>(&[
             (Ok(IA5String::new("abc").unwrap()), b"\x16\x03abc"),
             (Ok(IA5String::new(")").unwrap()), b"\x16\x01)"),
-            (Err(ParseError::InvalidValue), b"\x16\x03ab\xff"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x16\x03ab\xff",
+            ),
         ])
     }
 
@@ -504,7 +664,10 @@ mod tests {
         assert_parses::<Utf8String>(&[
             (Ok(Utf8String::new("abc")), b"\x0c\x03abc"),
             (Ok(Utf8String::new(")")), b"\x0c\x01)"),
-            (Err(ParseError::InvalidValue), b"\x0c\x01\xff"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x0c\x01\xff",
+            ),
         ])
     }
 
@@ -513,7 +676,10 @@ mod tests {
         assert_parses::<VisibleString>(&[
             (Ok(VisibleString::new("abc").unwrap()), b"\x1a\x03abc"),
             (Ok(VisibleString::new(")").unwrap()), b"\x1a\x01)"),
-            (Err(ParseError::InvalidValue), b"\x1a\x01\n"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x1a\x01\n",
+            ),
         ])
     }
 
@@ -552,30 +718,110 @@ mod tests {
                 Ok(UtcTime::new(Utc.ymd(1951, 5, 6).and_hms(23, 45, 0)).unwrap()),
                 b"\x17\x0b5105062345Z",
             ),
-            (Err(ParseError::InvalidValue), b"\x17\x0da10506234540Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d91a506234540Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d9105a6234540Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d910506a34540Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d910506334a40Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d91050633444aZ"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d910506334461Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e910506334400Za"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d000100000000Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d101302030405Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d100002030405Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d100100030405Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d100132030405Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d100231030405Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d100102240405Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d100102036005Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0d100102030460Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e-100102030410Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e10-0102030410Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e10-0002030410Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e1001-02030410Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e100102-030410Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e10010203-0410Z"),
-            (Err(ParseError::InvalidValue), b"\x17\x0e1001020304-10Z"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0da10506234540Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d91a506234540Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d9105a6234540Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d910506a34540Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d910506334a40Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d91050633444aZ",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d910506334461Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e910506334400Za",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d000100000000Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d101302030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d100002030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d100100030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d100132030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d100231030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d100102240405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d100102036005Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0d100102030460Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e-100102030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e10-0102030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e10-0002030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e1001-02030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e100102-030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e10010203-0410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0e1001020304-10Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x0c18102813516Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x17\x1018102813516+0730",
+            ),
         ]);
     }
 
@@ -604,34 +850,91 @@ mod tests {
                 )),
                 b"\x18\x1320100102030405-0607",
             ),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100102030405"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e00000100000000Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20101302030405Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100002030405Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100100030405Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100132030405Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100231030405Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100102240405Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100102036005Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0e20100102030460Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f-20100102030410Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f2010-0102030410Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f2010-0002030410Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f201001-02030410Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f20100102-030410Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f2010010203-0410Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f201001020304-10Z"),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100102030405",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e00000100000000Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20101302030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100002030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100100030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100132030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100231030405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100102240405Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100102036005Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0e20100102030460Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f-20100102030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f2010-0102030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f2010-0002030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f201001-02030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f20100102-030410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f2010010203-0410Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f201001020304-10Z",
+            ),
             // Tests for fractional seconds, which we currently don't support
             (
-                Err(ParseError::InvalidValue),
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
                 b"\x18\x1620100102030405.123456Z",
             ),
             (
-                Err(ParseError::InvalidValue),
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
                 b"\x18\x1520100102030405.123456",
             ),
-            (Err(ParseError::InvalidValue), b"\x18\x1020100102030405.Z"),
-            (Err(ParseError::InvalidValue), b"\x18\x0f20100102030405."),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x1020100102030405.Z",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x18\x0f20100102030405.",
+            ),
         ]);
     }
 
@@ -640,7 +943,7 @@ mod tests {
         assert_parses::<Enumerated>(&[
             (Ok(Enumerated::new(12)), b"\x0a\x01\x0c"),
             (
-                Err(ParseError::InvalidValue),
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
                 b"\x0a\x09\xff\xff\xff\xff\xff\xff\xff\xff\xff",
             ),
         ]);
@@ -653,9 +956,12 @@ mod tests {
                 Ok(Sequence::new(b"\x02\x01\x01\x02\x01\x02")),
                 b"\x30\x06\x02\x01\x01\x02\x01\x02",
             ),
-            (Err(ParseError::ShortData), b"\x30\x04\x02\x01\x01"),
             (
-                Err(ParseError::ExtraData),
+                Err(ParseError::new(ParseErrorKind::ShortData)),
+                b"\x30\x04\x02\x01\x01",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::ExtraData)),
                 b"\x30\x06\x02\x01\x01\x02\x01\x02\x00",
             ),
         ])
@@ -666,9 +972,12 @@ mod tests {
         assert_parses_cb(
             &[
                 (Ok((1, 2)), b"\x30\x06\x02\x01\x01\x02\x01\x02"),
-                (Err(ParseError::ShortData), b"\x30\x03\x02\x01\x01"),
                 (
-                    Err(ParseError::ExtraData),
+                    Err(ParseError::new(ParseErrorKind::ShortData)),
+                    b"\x30\x03\x02\x01\x01",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::ExtraData)),
                     b"\x30\x07\x02\x01\x01\x02\x01\x02\x00",
                 ),
             ],
@@ -688,7 +997,10 @@ mod tests {
                     b"\x30\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03",
                 ),
                 (Ok(vec![]), b"\x30\x00"),
-                (Err(ParseError::ShortData), b"\x30\x02\x02\x01"),
+                (
+                    Err(ParseError::new(ParseErrorKind::ShortData)),
+                    b"\x30\x02\x02\x01",
+                ),
             ],
             |p| {
                 p.read_element::<Sequence>()?.parse(|p| {
@@ -711,7 +1023,11 @@ mod tests {
                     b"\x30\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03",
                 ),
                 (Ok(vec![]), b"\x30\x00"),
-                (Err(ParseError::ShortData), b"\x30\x02\x02\x01"),
+                (
+                    Err(ParseError::new(ParseErrorKind::ShortData)
+                        .add_location(ParseLocation::Index(0))),
+                    b"\x30\x02\x02\x01",
+                ),
             ],
             |p| Ok(p.read_element::<SequenceOf<i64>>()?.collect()),
         )
@@ -727,12 +1043,20 @@ mod tests {
                 ),
                 (Ok(vec![]), b"\x31\x00"),
                 (
-                    Err(ParseError::InvalidSetOrdering),
+                    Err(ParseError::new(ParseErrorKind::InvalidSetOrdering)
+                        .add_location(ParseLocation::Index(1))),
                     b"\x31\x06\x02\x01\x03\x02\x01\x01",
                 ),
-                (Err(ParseError::ShortData), b"\x31\x01\x02"),
                 (
-                    Err(ParseError::UnexpectedTag { actual: 0x1 }),
+                    Err(ParseError::new(ParseErrorKind::ShortData)
+                        .add_location(ParseLocation::Index(0))),
+                    b"\x31\x01\x02",
+                ),
+                (
+                    Err(
+                        ParseError::new(ParseErrorKind::UnexpectedTag { actual: 0x1 })
+                            .add_location(ParseLocation::Index(0)),
+                    ),
                     b"\x31\x02\x01\x00",
                 ),
             ],
@@ -749,8 +1073,8 @@ mod tests {
                 (Ok((None, Some(18))), b"\x02\x01\x12"),
                 (Ok((Some(true), Some(18))), b"\x01\x01\xff\x02\x01\x12"),
                 (Ok((None, None)), b""),
-                (Err(ParseError::ShortData), b"\x01"),
-                (Err(ParseError::ShortData), b"\x02"),
+                (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x01"),
+                (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x02"),
             ],
             |p| {
                 Ok((
@@ -770,7 +1094,7 @@ mod tests {
                 b"\x04\x03abc",
             ),
             (Ok(None), b""),
-            (Err(ParseError::ShortData), b"\x04"),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x04"),
         ]);
 
         assert_parses::<Option<Choice2<u64, bool>>>(&[
@@ -784,8 +1108,13 @@ mod tests {
     fn test_choice1() {
         assert_parses::<Choice1<bool>>(&[
             (Ok(Choice1::ChoiceA(true)), b"\x01\x01\xff"),
-            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03\x00"),
-            (Err(ParseError::ShortData), b""),
+            (
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x03,
+                })),
+                b"\x03\x00",
+            ),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b""),
         ]);
     }
 
@@ -794,8 +1123,13 @@ mod tests {
         assert_parses::<Choice2<bool, i64>>(&[
             (Ok(Choice2::ChoiceA(true)), b"\x01\x01\xff"),
             (Ok(Choice2::ChoiceB(18)), b"\x02\x01\x12"),
-            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03\x00"),
-            (Err(ParseError::ShortData), b""),
+            (
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x03,
+                })),
+                b"\x03\x00",
+            ),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b""),
         ]);
     }
 
@@ -805,8 +1139,13 @@ mod tests {
             (Ok(Choice3::ChoiceA(true)), b"\x01\x01\xff"),
             (Ok(Choice3::ChoiceB(18)), b"\x02\x01\x12"),
             (Ok(Choice3::ChoiceC(())), b"\x05\x00"),
-            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03\x00"),
-            (Err(ParseError::ShortData), b""),
+            (
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x03,
+                })),
+                b"\x03\x00",
+            ),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b""),
         ]);
     }
 
@@ -817,11 +1156,15 @@ mod tests {
             (Ok(Implicit::new(true)), b"\x82\x01\xff"),
             (Ok(Implicit::new(false)), b"\x82\x01\x00"),
             (
-                Err(ParseError::UnexpectedTag { actual: 0x01 }),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x01,
+                })),
                 b"\x01\x01\xff",
             ),
             (
-                Err(ParseError::UnexpectedTag { actual: 0x02 }),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x02,
+                })),
                 b"\x02\x01\xff",
             ),
         ]);
@@ -830,14 +1173,18 @@ mod tests {
             (Ok(Implicit::new(Sequence::new(b"abc"))), b"\xa2\x03abc"),
             (Ok(Implicit::new(Sequence::new(b""))), b"\xa2\x00"),
             (
-                Err(ParseError::UnexpectedTag { actual: 0x01 }),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x01,
+                })),
                 b"\x01\x01\xff",
             ),
             (
-                Err(ParseError::UnexpectedTag { actual: 0x02 }),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x02,
+                })),
                 b"\x02\x01\xff",
             ),
-            (Err(ParseError::ShortData), b""),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b""),
         ]);
 
         assert_parses_cb(
@@ -845,8 +1192,14 @@ mod tests {
                 (Ok(Some(true)), b"\x82\x01\xff"),
                 (Ok(Some(false)), b"\x82\x01\x00"),
                 (Ok(None), b""),
-                (Err(ParseError::ExtraData), b"\x01\x01\xff"),
-                (Err(ParseError::ExtraData), b"\x02\x01\xff"),
+                (
+                    Err(ParseError::new(ParseErrorKind::ExtraData)),
+                    b"\x01\x01\xff",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::ExtraData)),
+                    b"\x02\x01\xff",
+                ),
             ],
             |p| p.read_optional_implicit_element::<bool>(2),
         );
@@ -855,8 +1208,14 @@ mod tests {
                 (Ok(Some(Sequence::new(b"abc"))), b"\xa2\x03abc"),
                 (Ok(Some(Sequence::new(b""))), b"\xa2\x00"),
                 (Ok(None), b""),
-                (Err(ParseError::ExtraData), b"\x01\x01\xff"),
-                (Err(ParseError::ExtraData), b"\x02\x01\xff"),
+                (
+                    Err(ParseError::new(ParseErrorKind::ExtraData)),
+                    b"\x01\x01\xff",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::ExtraData)),
+                    b"\x02\x01\xff",
+                ),
             ],
             |p| p.read_optional_implicit_element::<Sequence>(2),
         );
@@ -869,15 +1228,21 @@ mod tests {
             (Ok(Explicit::new(true)), b"\xa2\x03\x01\x01\xff"),
             (Ok(Explicit::new(false)), b"\xa2\x03\x01\x01\x00"),
             (
-                Err(ParseError::UnexpectedTag { actual: 0x01 }),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x01,
+                })),
                 b"\x01\x01\xff",
             ),
             (
-                Err(ParseError::UnexpectedTag { actual: 0x02 }),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x02,
+                })),
                 b"\x02\x01\xff",
             ),
             (
-                Err(ParseError::UnexpectedTag { actual: 0x03 }),
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x03,
+                })),
                 b"\xa2\x03\x03\x01\xff",
             ),
         ]);
@@ -887,9 +1252,14 @@ mod tests {
                 (Ok(Some(true)), b"\xa2\x03\x01\x01\xff"),
                 (Ok(Some(false)), b"\xa2\x03\x01\x01\x00"),
                 (Ok(None), b""),
-                (Err(ParseError::ExtraData), b"\x01\x01\xff"),
                 (
-                    Err(ParseError::UnexpectedTag { actual: 0x03 }),
+                    Err(ParseError::new(ParseErrorKind::ExtraData)),
+                    b"\x01\x01\xff",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                        actual: 0x03,
+                    })),
                     b"\xa2\x03\x03\x01\xff",
                 ),
             ],
