@@ -1,13 +1,17 @@
 use alloc::vec;
 use alloc::vec::Vec;
 use core::convert::TryInto;
+use core::hash::{Hash, Hasher};
 use core::marker::PhantomData;
 use core::mem;
 
 use chrono::{Datelike, TimeZone, Timelike};
 
 use crate::writer::Writer;
-use crate::{parse, parse_single, BitString, ObjectIdentifier, ParseError, ParseResult, Parser};
+use crate::{
+    parse, parse_single, BitString, ObjectIdentifier, ParseError, ParseErrorKind, ParseLocation,
+    ParseResult, Parser,
+};
 
 pub(crate) const CONTEXT_SPECIFIC: u8 = 0x80;
 pub(crate) const CONSTRUCTED: u8 = 0x20;
@@ -30,7 +34,9 @@ impl<'a, T: SimpleAsn1Readable<'a>> Asn1Readable<'a> for T {
     fn parse(parser: &mut Parser<'a>) -> ParseResult<Self> {
         let tlv = parser.read_tlv()?;
         if !Self::can_parse(tlv.tag) {
-            return Err(ParseError::UnexpectedTag { actual: tlv.tag });
+            return Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                actual: tlv.tag,
+            }));
         }
         Self::parse_data(tlv.data)
     }
@@ -68,7 +74,7 @@ impl<'a, T: SimpleAsn1Writable<'a>> SimpleAsn1Writable<'a> for &T {
 
 /// A TLV (type, length, value) represented as the tag and bytes content.
 /// Generally used for parsing ASN.1 `ANY` values.
-#[derive(Debug, PartialEq, PartialOrd, Clone, Copy)]
+#[derive(Debug, PartialEq, PartialOrd, Hash, Clone, Copy)]
 pub struct Tlv<'a> {
     pub(crate) tag: u8,
     // `data` is the value of a TLV
@@ -119,7 +125,7 @@ impl SimpleAsn1Readable<'_> for () {
         if data.is_empty() {
             Ok(())
         } else {
-            Err(ParseError::InvalidValue)
+            Err(ParseError::new(ParseErrorKind::InvalidValue))
         }
     }
 }
@@ -136,7 +142,7 @@ impl SimpleAsn1Readable<'_> for bool {
         match data {
             b"\x00" => Ok(false),
             b"\xff" => Ok(true),
-            _ => Err(ParseError::InvalidValue),
+            _ => Err(ParseError::new(ParseErrorKind::InvalidValue)),
         }
     }
 }
@@ -225,7 +231,8 @@ impl<'a> PrintableString<'a> {
 impl<'a> SimpleAsn1Readable<'a> for PrintableString<'a> {
     const TAG: u8 = 0x13;
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
-        PrintableString::new_from_bytes(data).ok_or(ParseError::InvalidValue)
+        PrintableString::new_from_bytes(data)
+            .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 
@@ -275,7 +282,7 @@ impl<'a> IA5String<'a> {
 impl<'a> SimpleAsn1Readable<'a> for IA5String<'a> {
     const TAG: u8 = 0x16;
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
-        IA5String::new_from_bytes(data).ok_or(ParseError::InvalidValue)
+        IA5String::new_from_bytes(data).ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 impl<'a> SimpleAsn1Writable<'a> for IA5String<'a> {
@@ -307,7 +314,8 @@ impl<'a> Utf8String<'a> {
 impl<'a> SimpleAsn1Readable<'a> for Utf8String<'a> {
     const TAG: u8 = 0x0c;
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
-        Utf8String::new_from_bytes(data).ok_or(ParseError::InvalidValue)
+        Utf8String::new_from_bytes(data)
+            .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 impl<'a> SimpleAsn1Writable<'a> for Utf8String<'a> {
@@ -361,7 +369,8 @@ impl<'a> VisibleString<'a> {
 impl<'a> SimpleAsn1Readable<'a> for VisibleString<'a> {
     const TAG: u8 = 0x1a;
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
-        VisibleString::new_from_bytes(data).ok_or(ParseError::InvalidValue)
+        VisibleString::new_from_bytes(data)
+            .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 impl<'a> SimpleAsn1Writable<'a> for VisibleString<'a> {
@@ -373,18 +382,18 @@ impl<'a> SimpleAsn1Writable<'a> for VisibleString<'a> {
 
 fn validate_integer(data: &[u8], signed: bool) -> ParseResult<()> {
     if data.is_empty() {
-        return Err(ParseError::InvalidValue);
+        return Err(ParseError::new(ParseErrorKind::InvalidValue));
     }
     // Ensure integer is minimally encoded
     if data.len() > 1
         && ((data[0] == 0 && data[1] & 0x80 == 0) || (data[0] == 0xff && data[1] & 0x80 == 0x80))
     {
-        return Err(ParseError::InvalidValue);
+        return Err(ParseError::new(ParseErrorKind::InvalidValue));
     }
 
     // Reject negatives for unsigned types.
     if !signed && data[0] & 0x80 == 0x80 {
-        return Err(ParseError::InvalidValue);
+        return Err(ParseError::new(ParseErrorKind::InvalidValue));
     }
 
     Ok(())
@@ -404,7 +413,7 @@ macro_rules! impl_asn1_element_for_int {
                     data = &data[1..];
                 }
                 if data.len() > mem::size_of::<Self>() {
-                    return Err(ParseError::IntegerOverflow);
+                    return Err(ParseError::new(ParseErrorKind::IntegerOverflow));
                 }
 
                 let mut fixed_data = [0; mem::size_of::<$t>()];
@@ -445,7 +454,7 @@ impl_asn1_element_for_int!(u64; false);
 /// Arbitrary sized unsigned integer. Contents may be accessed as `&[u8]` of
 /// big-endian data. Its contents always match the DER encoding of a value
 /// (i.e. they are minimal)
-#[derive(PartialEq, Clone, Copy, Debug)]
+#[derive(PartialEq, Clone, Copy, Debug, Hash)]
 pub struct BigUint<'a> {
     data: &'a [u8],
 }
@@ -469,7 +478,7 @@ impl<'a> BigUint<'a> {
 impl<'a> SimpleAsn1Readable<'a> for BigUint<'a> {
     const TAG: u8 = 0x02;
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
-        BigUint::new(data).ok_or(ParseError::InvalidValue)
+        BigUint::new(data).ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 impl<'a> SimpleAsn1Writable<'a> for BigUint<'a> {
@@ -482,7 +491,8 @@ impl<'a> SimpleAsn1Writable<'a> for BigUint<'a> {
 impl<'a> SimpleAsn1Readable<'a> for ObjectIdentifier<'a> {
     const TAG: u8 = 0x06;
     fn parse_data(data: &'a [u8]) -> ParseResult<ObjectIdentifier<'a>> {
-        ObjectIdentifier::from_der(data).ok_or(ParseError::InvalidValue)
+        ObjectIdentifier::from_der(data)
+            .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 impl<'a> SimpleAsn1Writable<'a> for ObjectIdentifier<'a> {
@@ -496,9 +506,10 @@ impl<'a> SimpleAsn1Readable<'a> for BitString<'a> {
     const TAG: u8 = 0x03;
     fn parse_data(data: &'a [u8]) -> ParseResult<BitString<'a>> {
         if data.is_empty() {
-            return Err(ParseError::InvalidValue);
+            return Err(ParseError::new(ParseErrorKind::InvalidValue));
         }
-        BitString::new(&data[1..], data[0]).ok_or(ParseError::InvalidValue)
+        BitString::new(&data[1..], data[0])
+            .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 impl<'a> SimpleAsn1Writable<'a> for BitString<'a> {
@@ -511,7 +522,7 @@ impl<'a> SimpleAsn1Writable<'a> for BitString<'a> {
 
 /// Used for parsing and writing ASN.1 `UTC TIME` values. Wraps a
 /// `chrono::DateTime<Utc>`.
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Debug, Clone, PartialEq, Hash)]
 pub struct UtcTime(chrono::DateTime<chrono::Utc>);
 
 impl UtcTime {
@@ -527,50 +538,37 @@ impl UtcTime {
     }
 }
 
-const UTCTIME_WITH_SECONDS_AND_OFFSET: &str = "%y%m%d%H%M%S%z";
-const UTCTIME_WITH_SECONDS: &str = "%y%m%d%H%M%SZ";
-const UTCTIME_WITH_OFFSET: &str = "%y%m%d%H%M%z";
-const UTCTIME: &str = "%y%m%d%H%MZ";
-
 impl SimpleAsn1Readable<'_> for UtcTime {
     const TAG: u8 = 0x17;
     fn parse_data(data: &[u8]) -> ParseResult<Self> {
-        let data = core::str::from_utf8(data).map_err(|_| ParseError::InvalidValue)?;
+        let data = core::str::from_utf8(data)
+            .map_err(|_| ParseError::new(ParseErrorKind::InvalidValue))?;
 
-        // Try parsing with every combination of "including seconds or not" and "fixed offset or
-        // UTC".
-        let mut result = None;
-        for format in [UTCTIME_WITH_SECONDS, UTCTIME].iter() {
-            if let Ok(dt) = chrono::Utc.datetime_from_str(data, format) {
-                result = Some(dt);
-                break;
-            }
+        // UTCTime comes in 4 different formats: with and without seconds, and
+        // with a fixed offset or UTC. We choose which to parse as based on the
+        // input length.
+        let mut dt = match data.len() {
+            17 => chrono::DateTime::parse_from_str(data, "%y%m%d%H%M%S%z").map(|dt| dt.into()),
+            15 => chrono::DateTime::parse_from_str(data, "%y%m%d%H%M%z").map(|dt| dt.into()),
+            13 => chrono::Utc.datetime_from_str(data, "%y%m%d%H%M%SZ"),
+            11 => chrono::Utc.datetime_from_str(data, "%y%m%d%H%MZ"),
+            _ => return Err(ParseError::new(ParseErrorKind::InvalidValue)),
         }
-        for format in [UTCTIME_WITH_SECONDS_AND_OFFSET, UTCTIME_WITH_OFFSET].iter() {
-            if let Ok(dt) = chrono::DateTime::parse_from_str(data, format) {
-                result = Some(dt.into());
-                break;
-            }
+        .map_err(|_| ParseError::new(ParseErrorKind::InvalidValue))?;
+        // Reject leap seconds, which aren't allowed by ASN.1. chrono encodes them as
+        // nanoseconds == 1000000.
+        if dt.nanosecond() >= 1_000_000 {
+            return Err(ParseError::new(ParseErrorKind::InvalidValue));
         }
-        match result {
-            Some(mut dt) => {
-                // Reject leap seconds, which aren't allowed by ASN.1. chrono encodes them as
-                // nanoseconds == 1000000.
-                if dt.nanosecond() >= 1_000_000 {
-                    return Err(ParseError::InvalidValue);
-                }
-                // UTCTime only encodes times prior to 2050. We use the X.509 mapping of two-digit
-                // year ordinals to full year:
-                // https://tools.ietf.org/html/rfc5280#section-4.1.2.5.1
-                if dt.year() >= 2050 {
-                    dt = chrono::Utc
-                        .ymd(dt.year() - 100, dt.month(), dt.day())
-                        .and_hms(dt.hour(), dt.minute(), dt.second());
-                }
-                Ok(UtcTime(dt))
-            }
-            None => Err(ParseError::InvalidValue),
+        // UTCTime only encodes times prior to 2050. We use the X.509 mapping of two-digit
+        // year ordinals to full year:
+        // https://tools.ietf.org/html/rfc5280#section-4.1.2.5.1
+        if dt.year() >= 2050 {
+            dt = chrono::Utc
+                .ymd(dt.year() - 100, dt.month(), dt.day())
+                .and_hms(dt.hour(), dt.minute(), dt.second());
         }
+        Ok(UtcTime(dt))
     }
 }
 
@@ -610,7 +608,7 @@ impl SimpleAsn1Writable<'_> for UtcTime {
 
 /// Used for parsing and writing ASN.1 `GENERALIZED TIME` values. Wraps a
 /// `chrono::DateTime<Utc>`.
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Clone, PartialEq, Hash)]
 pub struct GeneralizedTime(chrono::DateTime<chrono::Utc>);
 
 impl GeneralizedTime {
@@ -626,7 +624,8 @@ impl GeneralizedTime {
 impl SimpleAsn1Readable<'_> for GeneralizedTime {
     const TAG: u8 = 0x18;
     fn parse_data(data: &[u8]) -> ParseResult<GeneralizedTime> {
-        let data = core::str::from_utf8(data).map_err(|_| ParseError::InvalidValue)?;
+        let data = core::str::from_utf8(data)
+            .map_err(|_| ParseError::new(ParseErrorKind::InvalidValue))?;
         if let Ok(v) = chrono::Utc.datetime_from_str(data, "%Y%m%d%H%M%SZ") {
             return Ok(GeneralizedTime::new(v));
         }
@@ -634,7 +633,7 @@ impl SimpleAsn1Readable<'_> for GeneralizedTime {
             return Ok(GeneralizedTime::new(v.into()));
         }
 
-        Err(ParseError::InvalidValue)
+        Err(ParseError::new(ParseErrorKind::InvalidValue))
     }
 }
 
@@ -732,7 +731,7 @@ macro_rules! declare_choice {
                         return Ok($count::$name(tlv.parse::<$number>()?));
                     }
                 )*
-                Err(ParseError::UnexpectedTag{actual: tlv.tag()})
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag{actual: tlv.tag()}))
             }
 
             fn can_parse(tag: u8) -> bool {
@@ -830,16 +829,36 @@ impl<'a> SimpleAsn1Writable<'a> for SequenceWriter<'a> {
 /// are decoded.
 pub struct SequenceOf<'a, T: Asn1Readable<'a>> {
     parser: Parser<'a>,
+    length: usize,
     _phantom: PhantomData<T>,
 }
 
 impl<'a, T: Asn1Readable<'a>> SequenceOf<'a, T> {
     #[inline]
-    pub(crate) fn new(data: &'a [u8]) -> SequenceOf<'a, T> {
-        SequenceOf {
+    pub(crate) fn new(data: &'a [u8]) -> ParseResult<SequenceOf<'a, T>> {
+        let length = parse(data, |p| {
+            let mut i = 0;
+            while !p.is_empty() {
+                p.read_element::<T>()
+                    .map_err(|e| e.add_location(ParseLocation::Index(i)))?;
+                i += 1;
+            }
+            Ok(i)
+        })?;
+
+        Ok(SequenceOf {
+            length,
             parser: Parser::new(data),
             _phantom: PhantomData,
-        }
+        })
+    }
+
+    pub fn len(&self) -> usize {
+        self.length
+    }
+
+    pub fn is_empty(&self) -> bool {
+        self.len() == 0
     }
 }
 
@@ -847,22 +866,43 @@ impl<'a, T: Asn1Readable<'a>> Clone for SequenceOf<'a, T> {
     fn clone(&self) -> SequenceOf<'a, T> {
         SequenceOf {
             parser: self.parser.clone_internal(),
+            length: self.length,
             _phantom: PhantomData,
         }
     }
 }
 
+impl<'a, T: Asn1Readable<'a> + PartialEq> PartialEq for SequenceOf<'a, T> {
+    fn eq(&self, other: &Self) -> bool {
+        let mut it1 = self.clone();
+        let mut it2 = other.clone();
+        loop {
+            match (it1.next(), it2.next()) {
+                (Some(v1), Some(v2)) => {
+                    if v1 != v2 {
+                        return false;
+                    }
+                }
+                (None, None) => return true,
+                _ => return false,
+            }
+        }
+    }
+}
+
+impl<'a, T: Asn1Readable<'a> + Hash> Hash for SequenceOf<'a, T> {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        for val in self.clone() {
+            val.hash(state);
+        }
+    }
+}
+
 impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SequenceOf<'a, T> {
     const TAG: u8 = 0x10 | CONSTRUCTED;
     #[inline]
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
-        parse(data, |p| {
-            while !p.is_empty() {
-                p.read_element::<T>()?;
-            }
-            Ok(())
-        })?;
-        Ok(SequenceOf::new(data))
+        SequenceOf::new(data)
     }
 }
 
@@ -873,6 +913,7 @@ impl<'a, T: Asn1Readable<'a>> Iterator for SequenceOf<'a, T> {
         if self.parser.is_empty() {
             return None;
         }
+        self.length -= 1;
         Some(
             self.parser
                 .read_element::<T>()
@@ -938,6 +979,32 @@ impl<'a, T: Asn1Readable<'a>> Clone for SetOf<'a, T> {
     }
 }
 
+impl<'a, T: Asn1Readable<'a> + PartialEq> PartialEq for SetOf<'a, T> {
+    fn eq(&self, other: &Self) -> bool {
+        let mut it1 = self.clone();
+        let mut it2 = other.clone();
+        loop {
+            match (it1.next(), it2.next()) {
+                (Some(v1), Some(v2)) => {
+                    if v1 != v2 {
+                        return false;
+                    }
+                }
+                (None, None) => return true,
+                _ => return false,
+            }
+        }
+    }
+}
+
+impl<'a, T: Asn1Readable<'a> + Hash> Hash for SetOf<'a, T> {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        for val in self.clone() {
+            val.hash(state);
+        }
+    }
+}
+
 impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SetOf<'a, T> {
     const TAG: u8 = 0x11 | CONSTRUCTED;
 
@@ -945,15 +1012,21 @@ impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SetOf<'a, T> {
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
         parse(data, |p| {
             let mut last_element: Option<Tlv> = None;
+            let mut i = 0;
             while !p.is_empty() {
-                let el = p.read_tlv()?;
+                let el = p
+                    .read_tlv()
+                    .map_err(|e| e.add_location(ParseLocation::Index(i)))?;
                 if let Some(last_el) = last_element {
                     if el.full_data < last_el.full_data {
-                        return Err(ParseError::InvalidSetOrdering);
+                        return Err(ParseError::new(ParseErrorKind::InvalidSetOrdering)
+                            .add_location(ParseLocation::Index(i)));
                     }
                 }
                 last_element = Some(el);
-                el.parse::<T>()?;
+                el.parse::<T>()
+                    .map_err(|e| e.add_location(ParseLocation::Index(i)))?;
+                i += 1;
             }
             Ok(())
         })?;
@@ -1138,7 +1211,12 @@ impl<'a, T: Asn1Writable<'a>, const TAG: u8> SimpleAsn1Writable<'a> for Explicit
 
 #[cfg(test)]
 mod tests {
-    use crate::{parse_single, IA5String, ParseError, PrintableString, SequenceOf, Tlv};
+    use crate::{
+        parse_single, IA5String, ParseError, ParseErrorKind, PrintableString, SequenceOf, SetOf,
+        Tlv,
+    };
+    use std::collections::hash_map::DefaultHasher;
+    use std::hash::{Hash, Hasher};
 
     #[test]
     fn test_printable_string_new() {
@@ -1168,7 +1246,9 @@ mod tests {
         assert_eq!(tlv.parse::<u64>(), Ok(3));
         assert_eq!(
             tlv.parse::<&[u8]>(),
-            Err(ParseError::UnexpectedTag { actual: 0x2 })
+            Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                actual: 0x2
+            }))
         );
     }
 
@@ -1182,4 +1262,57 @@ mod tests {
         assert_eq!(seq1.collect::<Vec<_>>(), vec![2, 3]);
         assert_eq!(seq2.collect::<Vec<_>>(), vec![2, 3]);
     }
+
+    #[test]
+    fn test_sequence_of_len() {
+        let mut seq1 =
+            parse_single::<SequenceOf<u64>>(b"\x30\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03")
+                .unwrap();
+        let seq2 = seq1.clone();
+
+        assert_eq!(seq1.len(), 3);
+        assert!(seq1.next().is_some());
+        assert_eq!(seq1.len(), 2);
+        assert_eq!(seq2.len(), 3);
+        assert!(seq1.next().is_some());
+        assert!(seq1.next().is_some());
+        assert_eq!(seq1.len(), 0);
+        assert!(seq1.next().is_none());
+        assert_eq!(seq1.len(), 0);
+        assert!(seq1.is_empty());
+        assert_eq!(seq2.len(), 3);
+        assert!(!seq2.is_empty());
+    }
+
+    fn hash<T: Hash>(v: &T) -> u64 {
+        let mut h = DefaultHasher::new();
+        v.hash(&mut h);
+        h.finish()
+    }
+
+    #[test]
+    fn test_set_of_eq_hash() {
+        let s1 = SetOf::<bool>::new(b"");
+        let s2 = SetOf::<bool>::new(b"");
+        let s3 = SetOf::<bool>::new(b"\x01\x01\x00");
+
+        assert!(s1 == s2);
+        assert_eq!(hash(&s1), hash(&s2));
+
+        assert!(s2 != s3);
+        assert_ne!(hash(&s2), hash(&s3));
+    }
+
+    #[test]
+    fn test_sequence_of_eq_hash() {
+        let s1 = SequenceOf::<bool>::new(b"").unwrap();
+        let s2 = SequenceOf::<bool>::new(b"").unwrap();
+        let s3 = SequenceOf::<bool>::new(b"\x01\x01\x00").unwrap();
+
+        assert!(s1 == s2);
+        assert_eq!(hash(&s1), hash(&s2));
+
+        assert!(s2 != s3);
+        assert_ne!(hash(&s2), hash(&s3));
+    }
 }
